'''
The basic functions to disassemble pyc-files
'''

# Output and converting functions ------------------------------------------------------ 

import base64

def get_hex(code, prefix='', split=''):
    code = base64.b16encode(code).decode(encoding='UTF-8')
    if len(code)>0:
        code = split.join([(prefix+code[i:i + 2]) for i in range(0, len(code), 2)])
    return code    

def show_hex(code, label='', indent='', prefix='', split='', byteinline=16):
    '''
    Return string as:
    <indentlabel prefix00splitprefix01splitprefix10 ....>
    or
    <indentlabel
              prefix00splitprefix01splitprefix10 ....>
    if the code length is more than byteinline                     
    '''
    code = get_hex(code, prefix, split)
    maxlength = (2+len(split))*byteinline
    if len(code) < maxlength:
        if len(label)>0:
            print("%s%s %s" % (indent, label, code))
        else:
            print("%s%s" % (indent, code))    
    else:
        if len(label)>0:
            print("%s%s" % (indent, label))
            for i in range(0, len(code), maxlength):
                print("%s   %s" % (indent, code[i:i+maxlength]))
        else:
            for i in range(0, len(code), maxlength):
                print("%s%s" % (indent, code[i:i+maxlength]))            

# System analysis ----------------------------------------------------------------------

def get_python_version():   
    from imp import get_magic
    from struct import unpack
    from pythonversion import PythonVersionList
    magic = get_magic()
    try:
        number = unpack('<i', magic[:2]+b'\x00\x00')[0]
    except:
        number = 0
    if number not in PythonVersionList:
        number = 0
    return {'python': PythonVersionList[number], \
            'number': str(number), \
            'magic': get_hex(magic,'\\x')}

# File analysis ------------------------------------------------------------------------

def get_info_file(filename):
    '''
    Read and return file info 
    '''
    from marshal import load
    import time, struct
    try:
        f = open(filename,'rb')
        try:
            magic = f.read(4)
            moddate = f.read(4)
            modtime = time.asctime(time.localtime(struct.unpack('L', moddate)[0]))        
            try:
                code = load(f)
            except:
                f.seek(0)
                f.read(12)
                code = load(f)
        finally:
            f.close() 
    except Exception as E:
        print(str(E))
        return None
    else:
        return {'code': code, \
                'modtime': modtime, \
                'moddate': moddate, \
                'magic':magic}    

def get_bytecode_file(filename):
    info = get_info_file(filename)
    if info is not None:
        return info['code']
    else:
        return None

def get_compiler_file(filename):
    from imp import get_magic
    from struct import unpack
    from pythonversion import PythonVersionList
    magic = open(filename, "rb").read(4)
    try:
        number = unpack('<i', magic[:2]+b'\x00\x00')[0]
    except:
        number = 0
    if number not in PythonVersionList:
        number = 0
    return {'python': PythonVersionList[number], \
            'number': str(number), \
            'magic': get_hex(magic,'\\x')}    

def show_code(code, deep=True, indent='', level=0):
    '''
    Code analysis and print info. Syntax highlighting in Orion adapted to this format
    '''
    import types
    print("%s******************** code beginning (level %d) ********************" % (indent, level))
    storeindent = indent
    print("%sargcount:    %d" % (indent, code.co_argcount))
    print("%snlocals:     %d" % (indent, code.co_nlocals))
    print("%sstacksize:   %d" % (indent, code.co_stacksize))
    print("%sflags:       %04x" % (indent, code.co_flags))
    show_hex(code.co_code, "hexcode:",  indent, '', ' ')
    print("%sconsts:" % indent)
    for const in code.co_consts:
        print("   %s(%d) %r" % (indent, code.co_consts.index(const),const))
        if type(const) == types.CodeType and deep:
            show_code(const, deep, indent+'      ', level+1)
    print("%snames:       %r" % (indent, code.co_names))
    print("%svarnames:    %r" % (indent, code.co_varnames))
    print("%sfreevars:    %r" % (indent, code.co_freevars))
    print("%scellvars:    %r" % (indent, code.co_cellvars))
    print("%sfilename:    %r" % (indent, code.co_filename))
    print("%sname:        %r" % (indent, code.co_name))
    print("%sfirstlineno: %d" % (indent, code.co_firstlineno))
    show_hex(code.co_lnotab, "lnotab:", indent, '', ' ')
    print("%s********************** code ending (level %d) *********************" % (indent, level))
    
def show_file(filename):
    '''
    Print all info about pyc-file
    '''
    print("File: "+filename)
    print("------------------------------------------------------------------")
    info = get_info_file(filename)
    if info is not None:
        print("magic:   %s" % (get_hex(info["magic"],"\\x")))
        print("moddate: %s (%s)" % (get_hex(info["moddate"],"\\x"), info["modtime"]))
        show_code(info["code"])
    else:
        print("Check the compiler's version of this file!")

def show_compiler_file(filename):
    '''
    Print compiler's version of pyc-file
    '''
    print("File: "+filename)
    compiler = get_compiler_file(filename)
    print("Python: %s: %s (%s)" % (compiler["python"], compiler["number"], compiler["magic"]))
    
# Dis ----------------------------------------------------------------------------------

dis_out_buffer = ''
store_stdout = None
store_stderr = None

class DisStdErrHook:
    def write(self, message):
        pass
    def flush(self):
        pass
    def close(self):
        pass

class DisStdOutHook:
    def write(self, message):
        global dis_out_buffer
        if len(dis_out_buffer)>0:
            if dis_out_buffer[-1] == message == '\n':
                return
        dis_out_buffer += str(message)
    def flush(self):
        pass
    def close(self):
        pass

def dis_code(code, deep=True, indent='', level=0):
    '''
    Disassemble bytecode and print info. Syntax highlighting in Orion adapted to this format
    '''
    def set_hook():
        import sys
        global store_stdout, store_stderr, dis_out_buffer
        dis_out_buffer = ''
        store_stdout = sys.stdout
        sys.stdout = DisStdOutHook()
        #store_stderr = sys.stderr    #<<< off
        #sys.stderr = DisStdErrHook() #<<< off           
    def remove_hook():
        import sys
        global store_stdout, store_stderr
        #sys.stderr = store_stderr    #<<< off
        #store_stderr = None          #<<< off
        sys.stdout = store_stdout
        store_stdout = None
            
    import types
    global dis_out_buffer
    pyVer = get_python_version()
    if 'P3' in pyVer['python']: #FOR PYTHON 3.X
        import disassemble_3 as dis_module
        _have_code = (types.MethodType, types.FunctionType, types.CodeType, type)
        if code is None:
            return
        if hasattr(code, '__func__'): # Method
            code = code.__func__
        if hasattr(code, '__code__'): # Function
            code = code.__code__
        if hasattr(code, '__dict__'): # Class or module
            items = sorted(code.__dict__.items())
            for name, code1 in items:
                if isinstance(code1, _have_code) and deep:
                    print("# Disassembly of '%s':" % name)
                    try:
                        dis_code(code1, deep, indent, level)
                    except TypeError as msg:
                        print("Error:", msg)
        elif hasattr(code, 'co_code'): # Code object
            print("%s******************** code beginning (level %d) ********************" % (indent, level))
            print("%sdiscode[hexcode]:" % indent)
            set_hook()
            try:
                dis_module.disassemble(code)
            finally:
                remove_hook()
            if len(dis_out_buffer)>0:
                if dis_out_buffer[-1] == '\n':
                    dis_out_buffer = dis_out_buffer[:-1] #remove eol in text's end
                dis_out_buffer = ('%s|' % indent)+dis_out_buffer.replace('\n',('\n%s|' % indent))
                print(dis_out_buffer)
            if hasattr(code, 'co_consts'):
                print("%sconsts:" % indent)
                for const in code.co_consts:
                    print("   %s(%d) %r" % (indent, code.co_consts.index(const), const))
                    if type(const) == types.CodeType and deep:
                        dis_code(const, deep, indent+'      ', level+1)
            print("%s********************** code ending (level %d) *********************" % (indent, level))
        elif isinstance(code, (bytes, bytearray)): # Raw bytecode
            print("%s******************** code beginning (level %d) ********************" % (indent, level))
            print("%sdiscode[hexcode]:" % indent)
            set_hook()
            try:
                dis_module._disassemble_bytes(code)
            finally:
                remove_hook()
            if len(dis_out_buffer)>0:
                if dis_out_buffer[-1] == '\n':
                    dis_out_buffer = dis_out_buffer[:-1] #remove eol in text's end
                dis_out_buffer = ('%s|' % indent)+dis_out_buffer.replace('\n',('\n%s|' % indent))
                print(dis_out_buffer)
            print("%s********************** code ending (level %d) *********************" % (indent, level))
        elif isinstance(code, str):
            print("%s******************** code beginning (level %d) ********************" % (indent, level))
            print("%sdiscode[hexcode]:" % indent)
            set_hook()
            try:
                dis_module._disassemble_str(code)
            finally:
                remove_hook()
            if len(dis_out_buffer)>0:
                if dis_out_buffer[-1] == '\n':
                    dis_out_buffer = dis_out_buffer[:-1] #remove eol in text's end
                dis_out_buffer = ('%s|' % indent)+dis_out_buffer.replace('\n',('\n%s|' % indent))
                print(dis_out_buffer)
            print("%s********************** code ending (level %d) *********************" % (indent, level))
        else:
            print("Error: Don't know how to disassemble '%s' objects" % type(code).__name__)                        
    else: #FOR PYTHON 2.X
        import disassemble_2 as dis_module
        _have_code = (types.MethodType, types.FunctionType, types.CodeType, types.ClassType, type)
        if code is None:
            return
        if isinstance(code, types.InstanceType):
            code = code.__class__
        if hasattr(code, 'im_func'):
            code = code.im_func
        if hasattr(code, 'func_code'):
            code = code.func_code
        if hasattr(code, '__dict__'):
            items = code.__dict__.items()
            items.sort()
            for name, code1 in items:
                if isinstance(code1, _have_code) and deep:
                    print("# Disassembly of '%s':" % name)
                    try:
                        dis_code(code1, deep, indent, level)
                    except TypeError as msg:
                        print("Error:", msg)
        elif hasattr(code, 'co_code'):
            print("%s******************** code beginning (level %d) ********************" % (indent, level))
            print("%sdiscode[hexcode]:" % indent)
            set_hook()
            try:
                dis_module.disassemble(code)
            finally:
                remove_hook()
            if len(dis_out_buffer)>0:
                if dis_out_buffer[-1] == '\n':
                    dis_out_buffer = dis_out_buffer[:-1] #remove eol in text's end
                dis_out_buffer = ('%s|' % indent)+dis_out_buffer.replace('\n',('\n%s|' % indent))
                print(dis_out_buffer)
            if hasattr(code, 'co_consts'):
                print("%sconsts:" % indent)
                for const in code.co_consts:
                    print("   %s(%d) %r" % (indent, code.co_consts.index(const), const))
                    if type(const) == types.CodeType and deep:
                        dis_code(const, deep, indent+'      ', level+1)
            print("%s********************** code ending (level %d) *********************" % (indent, level))
        elif isinstance(code, str):
            print("%s******************** code beginning (level %d) ********************" % (indent, level))
            print("%sdiscode[hexcode]:" % indent)
            set_hook()
            try:
                dis_module.disassemble_string(code)
            finally:
                remove_hook()
            if len(dis_out_buffer)>0:
                if dis_out_buffer[-1] == '\n':
                    dis_out_buffer = dis_out_buffer[:-1] #remove eol in text's end
                dis_out_buffer = ('%s|' % indent)+dis_out_buffer.replace('\n',('\n%s|' % indent))
                print(dis_out_buffer)
            print("%s********************** code ending (level %d) *********************" % (indent, level))
        else:
            print("Error: Don't know how to disassemble '%s' objects" % type(code).__name__)

def dis_object(object, name=''):
    '''
    Disassemble bytecode of any Object
    '''
    print("Object: "+type(object).__name__+" "+("\'%s\'" % name if len(name)>0 else ""))
    print("------------------------------------------------------------------")
    dis_code(object)   
    
def dis_file(filename):
    '''
    Disassemble pyc-file's bytecode
    '''
    print("File: "+filename)
    print("------------------------------------------------------------------")
    info = get_info_file(filename)
    if info is not None:
        dis_code(info["code"])
    else:
        print("Check the compiler's version of this file!")